;;;;;; (not currently Meta-circular) Interpreter for Tunes' HLL- ;;;;;
;
; Modules needed:
;
(provide 'hll)
(require 'lib)		; all the functions
(require 'macrolib)	; all the macros
(require 'syntenv)	; syntactic environments
(require 'pattern)	; Pattern matching



;@ (Module HLL-Interpreter)


;;;;;;
;;;;;; Main evaluation function
;;;;;;

; Function HI:Val (HLL-interpreter evaluator) has 2 explicit parameters:
;
;  expr = expression to evaluate,
;  env  = environnent in which to evaluate.
;
; The result is (guess what ?)
; the evaluation of the expression in the given environment !
;
; It also has implicit parameters:
;  the meta-environment, which describes recognized object kinds:


; expressions are parsed syntax trees,
;	as given by read or quoted scheme expressions.
; environments are lists
; meta-environment is a list of (predicate action) lists.


(define HI:form-list (deq:new))

(define HI:meta-env HI:form-list)

(define (HI:get-form-list) (deq:->list HI:form-list))


;; form recognition structures:
(define (HI:fs:is-a-? fs) (car fs))
(define (HI:fs:eval-a fs) (cadr fs))
(define (HI:declare-form recognizer evaluator)
  (deq:push! HI:form-list (list recognizer evaluator)))


(define (HI:Val expr env)
  (list-find
    (lambda (x)		((HI:fs:is-a-? x) expr))
    (lambda (x r i l)   ((HI:fs:eval-a x) expr))
    (lambda (i l)	(ERROR "HLL-: form not recognized:" expr))
    (HI:get-form-list)))


(define (val expr env)
  (cond ((derived? expr)       (val-derived expr env))
        ((constant? expr)      (val-constant expr env))
        ((quote? expr)         (val-quote expr env))
        ((ref? expr)           (val-ref expr env))
        ((set? expr)           (val-set expr env))
        ((if? expr)            (val-if expr env))
        ((lambda? expr)        (val-lambda expr env))
        ((lambda-macro? expr)  (val-lambda-macro expr env)) ;;; constructeur de macros
        ((call? expr)          (val-call expr env))
	((define? expr)        (val-define expr env))       ;;; fonction de dfinition
        (else                  (error "erreur de syntaxe dans:" expr))))


(define (not-any? x)	       #f)
(define (val-error expr env)   (error "erreur de syntaxe dans:" expr))


;... in the end, put the following:

(map-2 HI:declare-form (list
	derived?	val-derived
	constant?	val-constant
	quote?		val-quote
	ref?		val-ref
	set?		val-set
	if?		val-if
	lambda?		val-lambda
	lambda-macro?	val-lambda-macro
	call?		val-call
	define?		val-define
	not-any?	val-error
	))


;;; ===========================================================================
;   Fonctions pour reconnaitre chaque forme d'expression.
;;; ===========================================================================

(define (constant? x) (or (boolean? x) (number? x) (char? x) (string? x)))
(define (ref? x)      (var? x))
(define (derived? x)  (and (non-empty-list? x) (transformable? (car x))))
(define (call? x)     (and (non-empty-list? x) (not (reserved? (car x)))))
(define (quote? x)    (and (non-empty-list? x) (eq? (car x) 'quote)))
(define (set? x)      (and (non-empty-list? x) (eq? (car x) 'set!)))
(define (if? x)       (and (non-empty-list? x) (eq? (car x) 'if)))
(define (lambda? x)   (and (non-empty-list? x) (eq? (car x) 'lambda)))
(define (lambda-macro? x)   (and (non-empty-list? x) (eq? (car x) 'lambda-macro))) ;;; nouveau
(define (define? x)   (and (non-empty-list? x) (eq? (car x) 'define)))             ;;; nouveau

(define (var? x)
  (and (symbol? x) (not (reserved? x))))

;;; Prdicat reconnaissant un paramtre formel pass par nom
(define (named-var? x)
  (and (list? x)
       (equal? (car x) 'name) 
       (var? (cadr x))
       (null? (cddr x))))

;;; Prdicat reconnaissant une liste de paramtres possiblement passs par nom.
(define (params? x)
  (cond ((null? x) #t)
        ((pair? x) (let ((element (car x))
			 (rest (cdr x)))
		     (and (or (var? element)
			      (named-var? element))
			  (params? rest))))
        (else      #f)))

;;; Prdicat reconnaissant une liste de paramtres qui ne sont pas passs nom.
(define (unnamed-params? x)
  (cond ((null? x) #t)
        ((pair? x) (let ((element (car x))
			 (rest (cdr x)))
		     (and (var? element)
			  (unnamed-params? rest))))
        (else      #f)))

;;; Prdicat reconnaissant les symboles rservs.
(define (reserved? x)
  (and (symbol? x)
       (or (member x '(quote set! if lambda lambda-macro define else =>)) ;;; lambda-macro nouveau
           (transformable? x))))

(define (non-empty-list? x)
  (and (pair? x) (proper-list? (cdr x))))

(define (proper-list? x) ; est-ce que le cdr de la derniere paire = () ?
  (or (null? x) (and (pair? x) (proper-list? (cdr x)))))

(define (proper-define? expr) ; est-ce une definition simple ou concise?
  (and (= (length expr) 3)
       (if (pair? (cadr expr))
         (and (var? (car (cadr expr)))
              (params? (cdr (cadr expr))))
         (var? (cadr expr)))))

(define (define-var expr) ; retourne la variable definie par un "define"
  (if (pair? (cadr expr))
    (car (cadr expr))
    (cadr expr)))

(define (define-expr expr) ; retourne l'expression contenue dans un "define"
  (if (pair? (cadr expr))
    (list 'lambda (cdr (cadr expr)) (caddr expr))
    (caddr expr)))

(define (all pred lst)
  (or (null? lst)
      (and (pred (car lst))
           (all pred (cdr lst)))))

(define (map1 proc lst)
  (if (null? lst)
    '()
    (cons (proc (car lst)) (map1 proc (cdr lst)))))

;;; ===========================================================================
;   Fonctions secondaires d'evaluation
;;; ===========================================================================


;;; ---------------------------------------------------------------------------
;;; Evalue une expression ayant define comme operateur. Ceci ajoute une
;;; liaison dans le frame courant de l'environnement. Retourne l'identificateur
;;; de la variable lie. Ce code provient de la boucle read-eval-print initiale.
;;; Define fonctionne donc correctement  tous les niveaux d'imbrication.
;;; Notre implantation des macro est facilite par cette modification.

(define (val-define expr env)
  (if (not (proper-define? expr))
      (error "`define' mal forme" expr)
      (begin
	(env-def (define-var expr)
		 (val (define-expr expr) env)
		 env)
	(define-var expr))))

(define (val-derived expr env)
  (val (transform expr) env))

(define (val-constant expr env)
  expr)

(define (val-quote expr env)
  (if (not (= (length expr) 2))
    (error "`quote' mal forme" expr)
    (cadr expr)))

(define (val-ref expr env)
  (env-fetch expr env))

(define (val-set expr env)
  (if (not (and (= (length expr) 3) (var? (cadr expr))))
    (error "`set!' mal forme" expr)
    (let ((value (val (caddr expr) env)))
      (begin
        (env-store (cadr expr) value env)
        (unspecified)))))

(define (unspecified)
  '*unspecified*) ; valeur arbitraire

(define (val-if expr env)
  (cond ((= (length expr) 3)
         (if (true? (val (cadr expr) env))
           (val (caddr expr) env)
           (unspecified)))
        ((= (length expr) 4)
         (if (true? (val (cadr expr) env))
           (val (caddr expr) env)
           (val (cadddr expr) env)))
        (else
         (error "`if' mal forme" expr))))

(define (true? x) ; en Scheme, tout ce qui est different de #f est "vrai"
  (not (eq? x #f)))

;;; ---------------------------------------------------------------------------
;;; L'valuation d'une lambda de notre sous-langage retourne une fermeture
;;; Scheme qui augmente l'environnement par les liaisons entre paramtres
;;; formels et actuels et qui evalue le corps de la lambda lorsqu'elle
;;; est appele. La lambda Scheme attend la liste de paramtres ainsi
;;; que l'environnement dynamique o les paramtres seront ventuellement
;;; valus. Ceux qui sont passs par valeur seront valus par
;;; make-frame-for-lambda (et donc avant l'evaluation du corps). Ceux qui
;;; sont passs par nom seront valus quand leur valeur sera requise
;;; dans l'valuation du corps.

(define (val-lambda expr lexical-env)
  (if (not (and (= (length expr) 3) (params? (cadr expr))))
    (error "`lambda' mal forme" expr)
    (lambda (args dynamic-env)
      (val (caddr expr)
           (env-extend (make-frame-for-lambda (cadr expr) args dynamic-env)
		       lexical-env)))))

;;; ---------------------------------------------------------------------------
;;; La fonction lambda-macro contruit une macro commme lambda construit une
;;; valeur fonctionelle. La valeur retourne est une fermeture Scheme qui,
;;;  la diffrence d'une fermeture cre par val-lambda, n'value pas les
;;; paramtres qui lui sont passs dans le premier argument. L'environnement
;;; dynamique pass  l'appel est donc inutilis. La corps de la lambda-macro
;;; doit s'valuer (dans l'environnement lexical)  une expression. Celle-ci
;;; sera ensuite value (dans l'environnement lexical encore une fois) pour
;;; obtenir la valeur de retour de l'application de la macro  ses paramtres.
;;; Les macros deviennent ainsi des objets de premire classe.
;;; La construction define-macro est elle mme une macro, dfinie  partir
;;; de lambda-macro.

(define (val-lambda-macro expr lexical-env)
  (if (not (and (= (length expr) 3) (unnamed-params? (cadr expr))))
    (error "`lambda-macro' mal forme" expr)
    (lambda (args dynamic-env)
      (val (val (caddr expr)
		(env-extend (make-frame-for-macro (cadr expr) args)
			    lexical-env))
	   lexical-env))))

(define (val-call expr env)
  (invoke (val (car expr) env)
          (cdr expr)
	  env))

(define (val-list lst env)
  (if (null? lst)
    '()
    (cons (val (car lst) env)
          (val-list (cdr lst) env))))

(define (invoke proc args env)
  (if (not (procedure? proc))
    (error "impossible d'appeler une non-procedure" proc)
    (proc args env)))

;;; ===========================================================================
;   Gestion des environnements
;;; ===========================================================================

;;; Determine si la longueur d'une liste indique un trio.
(define (trio? l) (= 3 (length l)))

(define (env-lookup var env) ; retourne une paire (ou #f si variable pas liee)
  (if (null? env)
    #f
    (let ((couple (assoc var (car env))))
      (if couple
        couple
        (env-lookup var (cdr env))))))

;;; ---------------------------------------------------------------------------
;;; Retourne la valeur  laquelle une variable est lie dans un environnement.
;;; Si l'entre dans l'environnement est une liste de deux lments, le
;;; deuxime est la valeur recherche. Si la liste  trois lments, elle
;;; provient d'un passage de paramtre par nom. Le deuxime lment est la
;;; valeur et doit tre valu dans l'environnement qui est le troisime
;;; lment. La valeur n'est donc plus contenue dans le cdr de la paire
;;; qui est maintenant une liste de 2 ou 3 lements.

(define (env-fetch var env)
  (let ((entry (env-lookup var env)))
    (if (not entry)
      (error "variable pas liee:" var)
      (if (trio? entry)
	  (val (cadr entry) (caddr entry))
	  (cadr entry)))))

;;; ---------------------------------------------------------------------------
;;; Assigne une nouvelle valeur  une variable lie dans un environnement.
;;; Si l'entre est une liste de 2 lments, on remplace le deuxime lment
;;; par la nouvelle valeur. Si la liste a trois lments, elle provient d'un 
;;; passage de paramtre par nom. Le deuxime lement doit tre une variable
;;; dont la valeur (dans l'environnement en troisime position) sera ensuite
;;; modifie.

(define (env-store var value env)
  (let ((entry (env-lookup var env)))
    (if (not entry)
      (error "variable pas liee:" var)
      (if (trio? entry)
	  (env-store (cadr entry) value (caddr entry))
	  (set-car! (cdr entry) value)))))

(define (env-extend env1 env2)
  (cons env1 env2))

(define (env-def var value env)
  (let ((entry (assoc var (car env))))
    (if entry
	(set-car! (cdr entry) value)
	(set-car! env (cons (list var value) (car env))))))

;;; ---------------------------------------------------------------------------
;;; Construit un frame qui peut tre rajout  un environnement  partir d'une liste de
;;; paramtres formels et une autre d'actuels, possiblement a
;;; valu dans un environnement.  Dans le cas ou les paramtres formels contiennent
;;; l'identificateur name, on ne forme pas un couple mais un trio.
;;; Le deuxime lment est l'expression non value et le troisime est l'environnent
;;; dans laquelle elle devra tre value lors des accs  la valeur de la variable.

(define (make-frame-for-lambda vars vals env)
  (cond ((null? vars)
         (if (null? vals)
           '()
           (error "trop de parametres actuels" "!")))
        ((null? vals)
         (error "pas assez de parametres actuels" "!"))
        (else
	 (let ((var (car vars))
	       (value (car vals))
	       (rest-vars (cdr vars))
	       (rest-vals (cdr vals)))
	   (cons (if (named-var? var)
		     (list (cadr var) value env)
		     (list var (val value env)))
		 (make-frame-for-lambda rest-vars rest-vals env))))))

;;; ---------------------------------------------------------------------------
;;; Construit un frame sans valuer les paramtres actuels auxquels
;;; les paramtres formels sont lis.

(define (make-frame-for-macro vars vals)
  (cond ((null? vars)
         (if (null? vals)
	     '()
	     (error "trop de parametres actuels" "!")))
        ((null? vals)
         (error "pas assez de parametres actuels" "!"))
	(else
         (cons (list (car vars) (car vals))
               (make-frame-for-macro (cdr vars) (cdr vals))))))


(define global-env (list '())) ; l'environnement d'evaluation global

;;; ===========================================================================
;   Implantation des formes derivees
;;; ===========================================================================

(define (transformable? head)
  (env-lookup head syntactic-env))

(define (transform expr)
  (let ((transformer (cadr (env-lookup (car expr) syntactic-env))))
    (transformer expr)))

(define syntactic-env (list '())) ; l'environnement de transformation

(define (transform-and expr)
  (cond ((null? (cdr expr))
         #t)
        ((pair? (cdr expr))
         (if (null? (cddr expr))
           (cadr expr)
           (list 'if (cadr expr) (cons 'and (cddr expr)) #f)))
        (else
         (error "`and' mal termine" expr))))

(env-def 'and transform-and syntactic-env)

(define (transform-or expr)
  (cond ((null? (cdr expr))
         #f)
        ((pair? (cdr expr))
         (if (null? (cddr expr))
           (cadr expr)
           (let ((temp (gensym))) ; variable temporaire
             (list 'let
                   (list (list temp (cadr expr)))
                   (list 'if temp temp (cons 'or (cddr expr)))))))
        (else
         (error "`or' mal termine" expr))))

(env-def 'or transform-or syntactic-env)

(define (transform-cond expr)
  (if (not (pair? (cdr expr)))

    (error "`cond' mal termine" expr)

    (let ((clause (cadr expr)))
      (if (not (non-empty-list? clause))

        (error "clause d'un `cond' mal terminee" clause)

        (if (eq? (car clause) 'else)

          (if (and (null? (cddr expr)) (not (null? (cdr clause))))
            (cons 'begin (cdr clause))
            (error "clause `else' d'un `cond' incorrecte" clause))

          (let ((temp (gensym)))
            (let ((consequent
                    (cond ((null? (cdr clause))
                           temp)
                          ((and (= (length clause) 3)
                                (eq? (cadr clause) '=>))
                           (list (caddr clause) temp))
                          (else
                           (cons 'begin (cdr clause))))))
              (list 'let
                    (list (list temp (car clause)))
                    (if (not (null? (cddr expr)))
                      (list 'if temp consequent (cons 'cond (cddr expr)))
                      (list 'if temp consequent))))))))))

(env-def 'cond transform-cond syntactic-env)

(define (transform-let expr)
  (if (or (not (= (length expr) 3))
          (not (proper-list? (cadr expr)))
          (not (all (lambda (x) (and (non-empty-list? x) (= (length x) 2)))
                    (cadr expr))))

    (error "`let' mal forme" expr)

    (cons (list 'lambda (map1 car (cadr expr)) (caddr expr))
          (map1 cadr (cadr expr)))))

(env-def 'let transform-let syntactic-env)

;;; ---------------------------------------------------------------------------
;;; Transformation du letrec
;;; Le letrec est transforme vers l'application d'une lambda, comme
;;; dans la tranformation d'un let. Les parametres passes ne sont
;;; cependant que des valeurs sans importance (#f). Le corps de la
;;; lambda generee contient les assignations necessaires pour
;;; donner les bonnes valeurs aux variables introduites par le letrec.
;;; Le corps du letrec est ensuite evalue a la fin du corps de la lambda.

(define (transform-letrec expr)
  (if (or (not (= (length expr) 3))
          (not (proper-list? (cadr expr)))
          (not (all (lambda (x) (and (non-empty-list? x) (= (length x) 2)))
                    (cadr expr))))

    (error "`letrec' mal forme")

    (let ((bindings (cadr expr))
          (formals (map1 car (cadr expr)))
          (body    (caddr expr)))
      (cons (list 'lambda formals
                  (list 'begin
                        (cons 'begin
                              (map1 (lambda (x) (cons 'set! x)) (cadr expr)))
                        body))
            (map1 (lambda (x) #f) bindings)))))

(env-def 'letrec transform-letrec syntactic-env)

(define (transform-begin expr)
  (if (not (pair? (cdr expr)))

    (error "`begin' mal termine" expr)

    (if (null? (cddr expr))
      (cadr expr)
      (let ((temp (gensym))) ; variable temporaire
        (list 'let
              (list (list temp (cadr expr)))
              (cons 'begin (cddr expr)))))))

(env-def 'begin transform-begin syntactic-env)

;;; ===========================================================================
;   Initialisation des variables predefinies
;;; ===========================================================================

;;; Les fermetures Scheme cres par ces fonctions reprsentent les fonctions de
;;; notre sous-langage et attendent donc un liste de paramtres et un
;;; environnement dans lequel elles doivent valuer ces paramtres.

(define (check-type-any name op)
  (lambda (args env) ;;; on passe l'environnement dynamique
    (if (= (length args) 1)
      (op (val (car args) env))  ;;; on value les paramtres dans l'env. dynamique
      (error "wrong number of arguments passed to" name))))

(define (check-type-any-any name op)
  (lambda (args env)
    (if (= (length args) 2)
      (op (val (car args) env) (val (cadr args) env))
      (error "wrong number of arguments passed to" name))))

(define (check-type-pair name op)
  (check-type-any
    name
    (lambda (x) (if (pair? x) (op x) (error "pair expected by" name)))))

(define (check-type-pair-any name op)
  (check-type-any-any
    name
    (lambda (x y) (if (pair? x) (op x y) (error "pair expected by" name)))))

(define (check-type-num-num name op)
  (check-type-any-any
    name
    (lambda (x y)
      (if (and (number? x) (number? y))
        (op x y)
        (error "number expected by" name)))))

(define (check-type-noarg name op)
  (lambda (args env) 
    (if (= (length args) 0)
      (op)
      (error "wrong number of arguments passed to" name))))



(define-macro


(map-macro
  (begin)
   register-scheme-primitive
     (null?       (any))
     (boolean?    (any))
     (number?     (any))
     (char?       (any))
     (string?     (any))
     (symbol?     (any))
     (pair?       (any))
     (list?       (any))
     (procedure?  (any))
     (eqv?        (any) (any))
     (eq?         (any) (any))
     (equal?      (any) (any))
     (not	  (any))
     (cons	  (any) (any))
     (car         (pair))
     (cdr         (pair))
     (set-car!    (pair) (any))
     (set-cdr!    (pair) (any))
     (length      (pair))
     (assoc	  (any) (pair))
     (member      (any) (pair))
     (<           (num) (num))
     (>           (num) (num))
     (<=          (num) (num))
     (>=          (num) (num))
     (=           (num) (num))
     (+           (num) (num))
     (-           (num) (num))
     (*           (num) (num))
     (/           (num) (num))
     (read)
     (write       (any))
     (list        (rest))
     ); end of registered primitives


;;; ----------------------------------------------------------------------
;;; define-macro est une variable contenant une macro (dfinie par macro).
;;; Cette macro reformule define-macro en termes de define et macro.
;;; define doit donc etre valuable. C'est pourquoi nous avons modifi la
;;; faon dont define est interpret.
;;; L'valuation de l'appel  lambda-macro ncessite videmment un appel
;;;  notre interprete. Cette definition peut aussi bien tre faite 
;;; l'incitateur de notre interprete.

(env-def 'define-macro (val '(lambda-macro (args body)
					   (list 'define (car args) 
						 (list 'lambda-macro (cdr args) 
						       body)))
			    global-env) 
	 global-env)


;;; -----------------------------------------------------------------------
;;; La boucle read-eval-print
;;; Une fois lance, on est entr dans un autre monde...

(let    ((rep-env  (initial-env)))
(letrec ((read-eval-print
    (write 'HLL>)
    (let* ((expr  (read))
           (value (HI:Val expr rep-env)))
    (write value))
    (newline)
    (read-eval-print)))

(read-eval-print)))
